import itertools
from functools import cached_property

import dask.array as da
import numpy as np
import skimage.measure
import skimage.transform

from palom import align, align_multi_res, img_util


def transform_bbox(bbox, affine_mx):
    tform_bbox = []
    tform = skimage.transform.AffineTransform(affine_mx)
    for rs, re, cs, ce in bbox:
        xx, yy = tform.inverse(
            list(itertools.product([cs, ce], [rs, re]))
        ).T
        rs2, cs2 = np.clip(np.floor([yy.min(), xx.min()]).astype(int), 0, None)
        re2, ce2 = np.ceil([yy.max(), xx.max()]).astype(int)
        tform_bbox.append([rs2, re2, cs2, ce2])
    return tform_bbox


class MultiObjAligner:

    def __init__(
        self,
        reader1, reader2,
        level1=0, level2=None,
        channel1=0, channel2=0,
        thumbnail_channel1=None, thumbnail_channel2=None,
        thumbnail_level1=-1,
    ) -> None:
        self.reader1 = reader1
        self.reader2 = reader2
        self.level1 = level1
        self.level2 = level2 or self._set_level2()

        self.channel1 = channel1
        self.channel2 = channel2
        self.thumbnail_channel1 = thumbnail_channel1 or channel1
        self.thumbnail_channel2 = thumbnail_channel2 or channel2
        self.thumbnail_level1 = thumbnail_level1

    def run(self, downscale_factor=8, exclude_objects=None):
        self.segment_objects(downscale_factor=downscale_factor, plot_segmentation=True)
        self.align_all_objects(plot_shift=True)
        self.combine_object_results(exclude_objects=exclude_objects)

    @cached_property
    def aligner(self):
        return self.make_aligner()

    @cached_property
    def ref_thumbnail(self):
        return np.array(self.aligner.ref_thumbnail)
    
    @cached_property
    def moving_thumbnail(self):
        return np.array(self.aligner.moving_thumbnail)
    
    @cached_property
    def fill_value_ref_thumbnail(self):
        return np.mean(
            self.ref_thumbnail[~img_util.entropy_mask(self.ref_thumbnail)]
        )
    
    @cached_property
    def fill_value_moving_thumbnail(self):
        return np.mean(
            self.moving_thumbnail[~img_util.entropy_mask(self.moving_thumbnail)]
        )
    
    @property
    def baseline_affine_matrix(self):
        if not hasattr(self, '_affine_matrix'):
            self._coarse_align()
        return self._affine_matrix

    @property
    def baseline_coarse_affine_matrix(self):
        if not hasattr(self, '_coarse_affine_matrix'):
            self._coarse_align()
        return self._coarse_affine_matrix
    
    @property
    def bbox_ref_thumbnail(self):
        if not hasattr(self, '_bbox_ref_thumbnail'):
            self.segment_objects(plot_segmentation=True)
        return self._bbox_ref_thumbnail
    
    @cached_property
    def bbox_ref_img_block(self):
        bbox = self.bbox_ref_thumbnail.astype(float)
        c21l = self.aligner
        downsample_factor = c21l.ref_thumbnail_down_factor
        rchunk, cchunk = c21l.ref_img.chunksize
        bbox[:, 0:2] *= downsample_factor / rchunk
        bbox[:, 2:4] *= downsample_factor / cchunk
        bbox[:, [0, 2]] = np.clip(np.floor(bbox[:, [0, 2]]), 0, None)
        bbox[:, [1, 3]] = np.ceil(np.ceil(bbox[:, [1, 3]]))
        return bbox.astype(int)

    def _coarse_align(self, **kwargs):
        c21l = self.make_aligner()
        default_kwargs = {
            'n_keypoints': 20_000,
            'plot_match_result': True,
            'test_flip': True,
            'test_intensity_invert': True,
            'auto_mask': True
        }
        c21l.coarse_register_affine(**{**default_kwargs, **kwargs})
        self._coarse_affine_matrix = c21l.coarse_affine_matrix
        self._affine_matrix = c21l.affine_matrix

    def segment_objects(self, downscale_factor=8, plot_segmentation=False):
        shape = self.ref_thumbnail.shape
        mask = img_util.entropy_mask(
            img_util.cv2_downscale_local_mean(
                self.ref_thumbnail, downscale_factor
            )
        )
        regionprops = skimage.measure.regionprops_table(
            skimage.morphology.label(mask),
            properties=['bbox', 'area']
        )
        bbox_ref_thumbnail = downscale_factor * np.array([
            regionprops['bbox-0'],
            regionprops['bbox-2'],
            regionprops['bbox-1'],
            regionprops['bbox-3']
        ]).T
        self._bbox_ref_thumbnail = bbox_ref_thumbnail[np.argsort(regionprops['area'])[::-1]]
        self.segmentation_mask = img_util.repeat_2d(
            skimage.morphology.label(mask), (downscale_factor, downscale_factor)
        )[:shape[0], :shape[1]]
        if plot_segmentation:
            self.plot_segmentation()

    def plot_segmentation(self):
        import matplotlib.pyplot as plt
        import matplotlib.patches
        import matplotlib.cm
        import skimage.segmentation
        colors = matplotlib.cm.Set3.colors
        fig, (ax1, ax2) = plt.subplots(1, 2)
        def _proc_img(img):
            if img_util.is_brightfield_img(img):
                return img
            return np.log1p(img)
        ax1.imshow(_proc_img(self.ref_thumbnail), cmap='gray')
        ax2.imshow(_proc_img(self.moving_thumbnail), cmap='gray')
        bounds = skimage.segmentation.find_boundaries(
            self.segmentation_mask, mode='thick'
        ).astype(float)
        ax1.imshow(np.where(bounds == 0, np.nan, bounds), cmap='cividis', vmin=0, vmax=1, interpolation='none')
        for idx, (rs, re, cs, ce) in enumerate(self.bbox_ref_thumbnail):
            color = colors[idx % len(colors)]
            mpatch = matplotlib.patches.Rectangle((cs, rs), ce-cs, re-rs, fill=False, edgecolor=color)
            ax1.add_patch(mpatch)

            corners = mpatch.get_corners()
            tform = skimage.transform.AffineTransform(self.baseline_coarse_affine_matrix)
            mpathc2 = matplotlib.patches.Polygon(tform.inverse(corners), fill=False, edgecolor=color)
            ax2.add_patch(mpathc2)
        return fig

    def _set_level2(self):
        lv_pairs = align_multi_res.match_levels(self.reader1, self.reader2)
        return lv_pairs[self.level1][1]
    
    def make_aligner(self):
        return align.get_aligner(
            self.reader1, self.reader2,
            level1=self.level1, level2=self.level2,
            channel1=self.channel1, channel2=self.channel2,
            thumbnail_level1=self.thumbnail_level1,
            thumbnail_level2=None,
            thumbnail_channel1=self.thumbnail_channel1,
            thumbnail_channel2=self.thumbnail_channel2,
        )
    
    def align_object(self, i, plot_shifts=True, **kwargs):
        rs, re, cs, ce = np.array(self.bbox_ref_thumbnail[i]).astype(int)
        rsm, rem, csm, cem = transform_bbox(
            self.bbox_ref_thumbnail, self.baseline_coarse_affine_matrix
        )[i]

        masked_t_ref = np.ones_like(self.ref_thumbnail) * self.fill_value_ref_thumbnail
        masked_t_ref[rs:re, cs:ce] = self.ref_thumbnail[rs:re, cs:ce]

        masked_t_moving = np.ones_like(self.moving_thumbnail) * self.fill_value_moving_thumbnail
        masked_t_moving[rsm:rem, csm:cem] = self.moving_thumbnail[rsm:rem, csm:cem]

        c21l = self.make_aligner()
        c21l.ref_thumbnail = masked_t_ref
        c21l.moving_thumbnail = masked_t_moving
        # FIXME image flipping should have been detected at this point
        default_kwargs = {
            'n_keypoints': 10_000,
            'plot_match_result': True,
            'test_flip': True,
            'test_intensity_invert': True,
            'auto_mask': True
        }
        c21l.coarse_register_affine(
            **{**default_kwargs, **kwargs}
        )
        if plot_shifts:
            import matplotlib.pyplot as plt
            plt.gcf().suptitle(f"Object {i} (coarse alignment)")

        rsf, ref, csf, cef = self.bbox_ref_img_block[i]
        shift_mask = da.zeros(c21l.grid_shape, dtype=bool, chunks=1)
        shift_mask[rsf:ref, csf:cef] = True

        c21l.compute_shifts(mask=shift_mask)
        if plot_shifts:
            try:
                c21l.plot_shifts()
                import matplotlib.pyplot as plt
                plt.gcf().suptitle(f"Object {i} (block shifts)")
            except Exception as e:
                print(f'\nFailed plotting shifts for object: {i}\n')
                print(e)
        c21l.constrain_shifts()
        return (c21l.shifts, c21l.block_affine_matrices, shift_mask)
    
    def align_all_objects(self, plot_shift=True):
        block_mxs = []
        shift_masks = []
        for idx, _ in enumerate(self.bbox_ref_thumbnail):
            _, mx, mask = self.align_object(idx, plot_shifts=plot_shift)
            block_mxs.append(mx)
            shift_masks.append(mask)
        self.block_mxs = np.array(block_mxs)
        self.shift_masks = np.array(shift_masks)

    def combine_object_results(self, exclude_objects=None):
        to_include = np.ones(len(self.shift_masks), dtype=bool)
        if exclude_objects is not None:
            for ii in exclude_objects:
                to_include[ii] = False
        assert to_include.sum() > 0
        masks = self.shift_masks[to_include]
        mxs = self.block_mxs[to_include]
        passed = np.argmax(
            masks.reshape(len(masks), -1), axis=0
        )
        mxs_final = np.zeros_like(mxs[0])
        for idx, bb in enumerate(mxs):
            mm = passed == idx
            mxs_final[mm] = bb[mm]
        mxs_final[
            ~masks.reshape(len(masks), -1).max(axis=0)
        ] = self.baseline_affine_matrix
        self.block_affine_matrices_da = align.block_affine_matrices_da(
            mxs_final, self.aligner.grid_shape
        )



import numpy as np

from . import align
from . import img_util


def match_levels(r1, r2):
    px_sizes_1 = [
        r1.pixel_size * r1.level_downsamples[i]
        for i in range(len(r1.pyramid))
    ]
    px_sizes_2 = [
        r2.pixel_size * r2.level_downsamples[i]
        for i in range(len(r2.pyramid))
    ]
    return [
        (idx, np.where(np.array(px_sizes_2) < px1)[0].max())
        if np.min(px_sizes_2) < px1
        else (idx, 0)
        for idx, px1 in enumerate(px_sizes_1)
    ]


class MultiResAligner:

    def __init__(
        self,
        reader1, reader2,
        level1=0,
        channel1=0, channel2=0,
        thumbnail_channel1=None, thumbnail_channel2=None,
        thumbnail_level1=-1,
        min_num_blocks=4
    ) -> None:
        self.reader1 = reader1
        self.reader2 = reader2
        self.level1 = level1
        
        self.channel1 = channel1
        self.channel2 = channel2
        self.thumbnail_channel1 = thumbnail_channel1 or channel1
        self.thumbnail_channel2 = thumbnail_channel2 or channel2
        self.thumbnail_level1 = thumbnail_level1

        self.min_num_blocks = min_num_blocks
        
        self._make_aligners()
    
    def __getstate__(self):
        state = self.__dict__.copy()
        props = ['aligners', 'base_aligners']
        for pp in props:
            if pp in state: del state[pp]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.__init__(
            state['reader1'], state['reader2'],
            level1=state['level1'],
            channel1=state['channel1'], channel2=state['channel2'],
            thumbnail_channel1=state['thumbnail_channel1'],
            thumbnail_channel2=state['thumbnail_channel2'],
            thumbnail_level1=state['thumbnail_level1'],
            min_num_blocks=state['min_num_blocks']
        )
        if '_coarse_affine_matrix' in state:
            for aligner in self.aligners:
                aligner.coarse_affine_matrix = state['_coarse_affine_matrix']
        if '_aligner_shifts' in state:
            if len(state['_aligner_shifts']) == len(self.aligners):
                for aa, ss in zip(self.aligners, state['_aligner_shifts']):
                    aa.shifts = ss
        # FIXME this does not fully recover the state before pickling, running
        # `constrain_shifts()` is required after loading the pickled

    @property
    def levels(self):
        levels = filter(
            lambda x: x >= self.level1,
            self.reader1.level_downsamples.keys()
        )
        levels = filter(
            lambda x: np.prod(self.reader1.pyramid[x].numblocks) >= self.min_num_blocks,
            levels
        )
        return sorted(levels)

    @property
    def downsample_factors(self):
        return [
            self.reader1.level_downsamples[ll]
            for ll in self.levels
        ]

    @property
    def level_pairs(self):
        levels = match_levels(self.reader1, self.reader2)
        filtered = filter(lambda x: x[0] >= self.level1, levels)
        return list(filtered)
    
    @property
    def base_aligner(self):
        import copy
        aligner = copy.deepcopy(self.aligners[0])
        if hasattr(self, 'shifts'):
            aligner.shifts = self.shifts
        return aligner

    @property
    def coarse_affine_matrix(self):
        if not hasattr(self, '_coarse_affine_matrix'):
            self._coarse_align()
        return self._coarse_affine_matrix
    
    def _make_aligners(self):
        self.level2 = self.level_pairs[0][1]
        self.aligners = []
        for l1, l2 in self.level_pairs:
            c21l = align.get_aligner(
                self.reader1, self.reader2, 
                channel1=self.channel1, channel2=self.channel2,
                level1=l1, level2=l2,
                thumbnail_channel1=self.thumbnail_channel1,
                thumbnail_channel2=self.thumbnail_channel2,
                thumbnail_level1=self.thumbnail_level1,
                # FIXME handle user selected thumbnail level
                thumbnail_level2=None
            )
            if c21l.num_blocks < self.min_num_blocks:
                continue
            self.aligners.append(c21l)

    def _coarse_align(self, **kwargs):
        l1, l2 = self.level_pairs[0]
        aligner = align.get_aligner(
            self.reader1, self.reader2,
            thumbnail_channel1=self.thumbnail_channel1,
            thumbnail_channel2=self.thumbnail_channel2,
            thumbnail_level1=self.thumbnail_level1,
            # FIXME handle user selected thumbnail level
            thumbnail_level2=None,
            channel1=self.channel1,
            channel2=self.channel2,
            level1=l1, level2=l2,
        )
        default_kwargs = {
            'n_keypoints': 20_000,
            'plot_match_result': True,
            'test_flip': True,
            'test_intensity_invert': True,
        }
        aligner.coarse_register_affine(**{**default_kwargs, **kwargs})
        self._coarse_affine_matrix = aligner.coarse_affine_matrix

    def align(self):
        self._aligner_shifts = []
        for aligner in self.aligners:
            aligner.coarse_affine_matrix = self.coarse_affine_matrix
            aligner.compute_shifts()
            self._aligner_shifts.append(aligner.shifts)

    def constrain_shifts(self, exclude_result_levels=None):
        aligners = self.aligners
        for aligner in aligners:
            # FIXME workaround to manually exclude computed shifts from certain levels
            if not hasattr(aligner, 'original_shifts'):
                aligner.constrain_shifts()
        _valid_masks = [
            np.all(al.original_shifts == al.shifts, axis=1)
            for al in aligners
        ]
        h, w = aligners[0].grid_shape
        downsample_factors = [
            int(dd / self.downsample_factors[0])
            for dd in self.downsample_factors
        ]
        valid_masks = [
            img_util.repeat_2d(mm.reshape(aa.grid_shape), (dd, dd))[:h, :w]
            for dd, mm, aa in zip(
                downsample_factors, _valid_masks, aligners
            ) 
        ]
        idxs = [
            img_util.repeat_2d(
                np.arange(aa.shifts.shape[0]).reshape(aa.grid_shape), 
                (dd, dd)
            )[:h, :w]
            for aa, dd in zip(aligners, downsample_factors)
        ]
        exclude_result_levels = exclude_result_levels or []
        for level in exclude_result_levels:
            valid_masks[level][:] = False
        mask = np.argmax(valid_masks, axis=0)
        out = np.zeros((2, *aligners[0].grid_shape))
        for ii, (aa, idx, dd) in enumerate(
            zip(aligners, idxs, downsample_factors)
        ):
            out[[mask == ii]*2] = (
                dd * aa.shifts[idx[mask == ii]].T.flatten()
            )
        self.shifts = out.reshape(2, -1).T
        self.valid_masks = valid_masks
        self.idxs = idxs

    def plot_shifts(self, max_radius=None):
        import matplotlib.pyplot as plt
        import matplotlib.figure
        import skimage.color
        from mpl_toolkits.axes_grid1 import make_axes_locatable

        from .cli import flow
        
        shape = self.base_aligner.grid_shape
        shifts = self.shifts.T.reshape(2, *shape)

        shift_mask = shifts
        if hasattr(self.base_aligner, 'original_shifts'):
            shift_mask = self.base_aligner.original_shifts.T.reshape(2, *shape)
        shift_mask = np.all(np.isfinite(shift_mask), axis=0)
        mask = np.max(self.valid_masks, axis=0) & shift_mask

        if max_radius is None:
            max_radius = np.percentile(np.linalg.norm(shifts, axis=0)[mask], 99.5)

        lab = flow.shifts_to_lab(
            shifts,
            max_radius=max_radius
        )
        rgb = skimage.color.lab2rgb(lab, channel_axis=0)
        
        thumbnail = self.reader1.pyramid[-1][self.thumbnail_channel1].compute()
        thumbnail_extent = flow.get_img_extent(
            thumbnail.shape,
            self.reader1.level_downsamples[len(self.reader1.pyramid)-1]
        )
        
        flow_extent = flow.get_img_extent(
            shape,
            self.downsample_factors[0] * self.reader1.pyramid[self.levels[0]].chunksize[1]
        )

        w, h = matplotlib.figure.figaspect(shape[0] / (shape[1] * 3))
        fig = plt.figure(figsize=(w, h))
        gs = fig.add_gridspec(1, 3, width_ratios=(.5, 1, 1))

        ax1 = fig.add_subplot(gs[1])
        ax1.imshow(np.log1p(thumbnail), alpha=1, extent=thumbnail_extent, cmap='gray')
        ax1.imshow(np.dstack([*rgb, mask]), extent=flow_extent, alpha=0.8)
        _cax = make_axes_locatable(ax1).append_axes("right", size="5%", pad=0.05)
        _cax.axis('off')
        ax1.axis('off')

        ax2 = fig.add_subplot(gs[2], sharex=ax1, sharey=ax1)
        ax2.imshow(np.log1p(thumbnail), alpha=1, extent=thumbnail_extent, cmap='gray')
        im = ax2.imshow(
            np.argmax(self.valid_masks, axis=0),
            extent=flow_extent, alpha=0.5,
            cmap='Set3', vmin=-.5, vmax=12-.5,
        )
        divider = make_axes_locatable(ax2)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(im, cax=cax, ticks=self.levels, values=self.levels)

        _ = flow.plot_legend(
            np.array([*shifts, mask]),
            max_radius, True, True, plot_flow=False, ax=fig.add_subplot(gs[0])
        )
        return fig



import pathlib

import tifffile
from loguru import logger

import palom


PYRAMID_DEFAULTS = dict(
    downscale_factor=2,
    compression='zlib',
    tile_size=1024,
    save_RAM=True
)


def validate_out_path(out_path, default, overwrite):
    if out_path is None:
        out_path = default
    out_path = pathlib.Path(out_path).absolute()
    if ''.join(out_path.suffixes[-2:]).lower() not in ('.ome.tif', '.ome.tiff'):
        logger.error(
            'out_path must ends with .ome.tif or .ome.tiff'
        )
    if (not overwrite) & out_path.exists():
        logger.error(
            f"Aborted. Destination file exists {out_path.name}."
            " Set `overwrite=True` to overwrite it."
        )
        raise OSError('File exists')
    return out_path


def merge_channels(img_paths, out_path=None, overwrite=False, pyramid_config=None):
    readers = [palom.reader.OmePyramidReader(pp) for pp in img_paths]
    img_path = pathlib.Path(img_paths[0])
    stem = img_path.name.replace(''.join(img_path.suffixes), '')
    out_path = validate_out_path(
        out_path,
        img_path.parent / f"merged-{stem}-zlib.ome.tif",
        overwrite=overwrite
    )
    mosaics = [rr.pyramid[0] for rr in readers]
    pixel_size = readers[0].pixel_size
    try:
        tif_tags = src_tif_tags(img_path)
    except Exception:
        tif_tags = {}
    if pyramid_config is None: pyramid_config = {}
    text = '''
    Processing:
        {}
    '''.format('\n\t'.join([pathlib.Path(p).name for p in img_paths]))
    logger.info(text)
    palom.pyramid.write_pyramid(
        mosaics,
        out_path,
        **{
            **dict(
                pixel_size=pixel_size,
                kwargs_tifffile=tif_tags,
            ),
            **PYRAMID_DEFAULTS,
            **pyramid_config
        }
    )
    return out_path


def extract_channels(
    img_path, channels, out_path=None, overwrite=False, pyramid_config=None
):
    img_path = pathlib.Path(img_path)
    reader = palom.reader.OmePyramidReader(img_path)
    stem = img_path.name.replace(''.join(img_path.suffixes), '')
    out_path = validate_out_path(
        out_path,
        img_path.parent / f"extracted-{stem}-channel-{'_'.join([str(c) for c in channels])}.ome.tif",
        overwrite=overwrite
    )
    mosaics = [reader.pyramid[0][channels]]
    try:
        tif_tags = src_tif_tags(img_path)
    except Exception:
        tif_tags = {}
    if pyramid_config is None: pyramid_config = {}
    text = f'''
    Processing: {img_path}
        extract channels: {channels}
    '''
    logger.info(text)
    palom.pyramid.write_pyramid(
        mosaics,
        out_path,
        **{
            **dict(
                pixel_size=reader.pixel_size,
                kwargs_tifffile=tif_tags,
            ),
            **PYRAMID_DEFAULTS,
            **pyramid_config
        }
    )
    return out_path


def src_tif_tags(img_path):
    kwargs_tifffile = {}
    with tifffile.TiffFile(img_path) as tif:
        kwargs_tifffile.update(dict(
            photometric=tif.pages[0].photometric.value,
            resolution=tif.pages[0].resolution,
            resolutionunit=tif.pages[0].resolutionunit.value,
            software=tif.pages[0].software
        ))
    return kwargs_tifffile


def compress_pyramid(img_path, out_path=None, overwrite=False):
    img_path = pathlib.Path(img_path)
    out_path = merge_channels([img_path], out_path, overwrite)
    ome_xml = tifffile.tiffcomment(img_path)
    tifffile.tiffcomment(out_path, ome_xml.encode())
    return out_path


def fix_stitcher_ome_xml(
    img_path,
    replace_from,
    replace_to
):
    ori = tifffile.tiffcomment(img_path)
    n_to_replace = ori.count(replace_from)
    if n_to_replace == 0:
        print(f"Substring to be replaced not found in the file ({img_path})")
        return 0
    fixed = ori.replace(replace_from, replace_to)
    tifffile.tiffcomment(img_path, fixed.encode())
    print(f"{n_to_replace} instance(s) of {replace_from} replaced with {replace_to}")
    return 0


def compress_rarecyte_ome_tiff(img_path, out_path=None, overwrite=False):
    ome_path = compress_pyramid(img_path, out_path, overwrite)
    fix_stitcher_ome_xml(
        ome_path,
        '</Channel><Plane',
        '</Channel><MetadataOnly></MetadataOnly><Plane'
    )
    return


def main():
    import fire
    fire.Fire({
        'merge': merge_channels,
        'extract': extract_channels,
        'compress': compress_pyramid,
        'compress-rarecyte': compress_rarecyte_ome_tiff,
    })


if __name__ == '__main__':
    import sys
    sys.exit(main())


import argparse
import pathlib
import sys
import yaml


def main(argv=sys.argv):

    parser = argparse.ArgumentParser(
        description=(
            'Generate configuration yaml file for palom-svs run'
        )
    )
   
    parser.add_argument(
        '-i',
        metavar='input-dir',
        help='directory that contains all the SVS files to be processed',
        required=True
    )
    parser.add_argument(
        '-n',
        metavar='ref-name-pattern',
        help='filename pattern of the reference SVS file',
        required=True
    )
    parser.add_argument(
        '-o',
        metavar='output-ome-path',
        help='full path to the output ome-tiff file',
        required=True
    )
    parser.add_argument(
        '-c',
        metavar='yaml-path',
        help='full path to the resulting configuration yaml file',
        required=True
    )
   
    args = parser.parse_args(argv[1:])
   
    if len(argv) == 1:
        parser.print_help()
        return 0

    svs_dir = pathlib.Path(args.i)
    name_pattern = args.n
    output_path = pathlib.Path(args.o)
    yaml_path = pathlib.Path(args.c)

    svs_paths = sorted(svs_dir.glob('*.svs'))
    svs_paths.sort(key=lambda x: x.name.split('_')[-2])

    assert len(svs_paths) > 0
    assert '.ome.tif' in str(output_path)
    assert '.yml' in str(yaml_path) or '.yaml' in str(yaml_path)

    ref_slide = None
    for svs in svs_paths:
        if svs.match(name_pattern):
            print('Reference file found {}'.format(svs.name))
            ref_slide = svs
            svs_paths.remove(svs)
            break

    assert ref_slide is not None, (
        f"Cannot find reference SVS file with name pattern: {name_pattern}"
    )

    channel_names = [
        '-'.join(p.name.split('_')[-2:][::-1]).replace('.svs', '')
        for p in ([ref_slide] + svs_paths)
    ]

    config = {
        'input dir': str(svs_dir),
        'output full path': str(output_path),
        'reference image': {
            'filename': str(ref_slide).replace(str(svs_dir), '.'),
            'output mode': 'hematoxylin',
            'channel name': channel_names[0]
        },
        'moving images': [
            {
                'filename': str(s).replace(str(svs_dir), '.'),
                'output mode': 'aec' if not s.match(name_pattern) else 'hematoxylin',
                'channel name': n
            }
            for s, n in zip(svs_paths, channel_names[1:])
        ]
    }

    with open(yaml_path, 'w') as f:
        yaml.dump(config, f, sort_keys=False)


if __name__ == '__main__':
    sys.exit(main())



import pathlib

import matplotlib
import matplotlib.pyplot as plt

import palom


def align_he(
    p1: str | pathlib.Path,
    p2: str | pathlib.Path,
    out_dir: str | pathlib.Path,
    out_name: str = None,
    thumbnail_channel1: int = 1,
    thumbnail_channel2: int = 2,
    channel1: int = 0,
    channel2: int = 2,
    px_size1: float = None,
    px_size2: float = None,
    n_keypoints: int = 10_000,
    auto_mask=True,
    only_coarse: bool = False,
    only_qc: bool = False,
    viz_coarse_napari: bool = False,
    multi_res: bool = False,
    multi_obj: bool = False,
    multi_obj_kwarg: dict = None,
):
    assert not (multi_res and multi_obj), (
        "setting both `multi_res` and `multi_obj` to `True` is not supported,"
        " choose at most one"
    )
    out_dir, p1, p2 = pathlib.Path(out_dir), pathlib.Path(p1), pathlib.Path(p2)
    if out_name is None:
        out_name = f"{p2.stem}-registered.ome.tif"
    out_path = out_dir / out_name
    assert "".join(out_path.suffixes[-2:]) in (".ome.tif", ".ome.tiff")
    out_path.parent.mkdir(exist_ok=True, parents=True)
    set_matplotlib_font(font_size=8)

    r1 = get_reader(p1)(p1, pixel_size=px_size1)
    r2 = get_reader(p2)(p2, pixel_size=px_size2)

    LEVEL = 0
    aligner = palom.align.get_aligner(
        r1,
        r2,
        level1=LEVEL,
        level2=LEVEL,
        channel1=channel1,
        channel2=channel2,
        # make thumbnail level pair based on pixel_size
        thumbnail_level1=None,
        thumbnail_channel1=thumbnail_channel1,
        thumbnail_channel2=thumbnail_channel2,
    )

    aligner.coarse_register_affine(
        n_keypoints=n_keypoints,
        test_flip=True,
        test_intensity_invert=True,
        auto_mask=auto_mask,
    )
    fig, ax = plt.gcf(), plt.gca()
    fig.suptitle(f"{p2.name} (coarse alignment)", fontsize=8)
    ax.set_title(f"{p1.name} - {p2.name}", fontsize=6)
    im_h, im_w = ax.images[0].get_array().shape
    set_subplot_size(im_w / 288, im_h / 288, ax=ax)
    ax.set_anchor("N")
    # use 0.5 inch on the top for figure title
    fig.subplots_adjust(top=1 - 0.5 / fig.get_size_inches()[1])
    save_all_figs(out_dir=out_dir / "qc", format="jpg", dpi=144)

    if viz_coarse_napari:
        _ = viz_coarse(r1, r2, LEVEL, LEVEL, channel1, channel2, aligner.affine_matrix)

    if not only_coarse:
        # the default
        if not (multi_res or multi_obj):
            aligner.compute_shifts()

            fig = aligner.plot_shifts()
            fig.suptitle(f"{p2.name} (block shift distance)", fontsize=8)
            fig.axes[0].set_title(p1.name, fontsize=6)
            save_all_figs(out_dir=out_dir / "qc", format="png")

            aligner.constrain_shifts()
            block_mx = aligner.block_affine_matrices_da
        elif multi_res:
            mr_aligner = palom.align_multi_res.MultiResAligner(
                r1,
                r2,
                level1=LEVEL,
                channel1=channel1,
                channel2=channel2,
                thumbnail_channel1=thumbnail_channel1,
                thumbnail_channel2=thumbnail_channel2,
                min_num_blocks=25,
            )
            mr_aligner._coarse_affine_matrix = aligner.coarse_affine_matrix
            mr_aligner.align()
            mr_aligner.constrain_shifts()

            fig = mr_aligner.plot_shifts()
            fig.suptitle(f"{p2.name} (multi-res aligment)", fontsize=8)
            fig.axes[0].set_title(p1.name, fontsize=6)
            save_all_figs(out_dir=out_dir / "qc", format="png", dpi=144)

            pickle_dir = out_dir / "pickle"
            if not pickle_dir.exists():
                pickle_dir.mkdir(parents=True)
            import pickle

            with open(pickle_dir / f"{p2.name}-palom.pkl", "wb") as f:
                pickle.dump(mr_aligner, f)

            aligner = mr_aligner.base_aligner
            block_mx = mr_aligner.base_aligner.block_affine_matrices_da
        elif multi_obj:
            mo_aligner = palom.align_multi_obj.MultiObjAligner(
                r1,
                r2,
                level1=LEVEL,
                channel1=channel1,
                channel2=channel2,
                thumbnail_channel1=thumbnail_channel1,
                thumbnail_channel2=thumbnail_channel2,
            )
            mo_aligner._affine_matrix = aligner.affine_matrix
            mo_aligner._coarse_affine_matrix = aligner.coarse_affine_matrix
            if multi_obj_kwarg is None:
                multi_obj_kwarg = {}
            mo_aligner.run(**multi_obj_kwarg)
            save_all_figs(
                out_dir=out_dir / "qc" / p2.stem, format="png", dpi=144, prefix=p2.name
            )
            block_mx = mo_aligner.block_affine_matrices_da

    if not only_qc:
        mx = aligner.affine_matrix
        if not only_coarse:
            mx = block_mx

        mosaic = palom.align.block_affine_transformed_moving_img(
            ref_img=aligner.ref_img, moving_img=r2.pyramid[LEVEL], mxs=mx
        )
        palom.pyramid.write_pyramid(
            mosaics=[mosaic],
            output_path=out_path,
            pixel_size=r1.pixel_size * r1.level_downsamples[LEVEL],
            channel_names=[list("RBG")],
            compression="zlib",
            downscale_factor=2,
            save_RAM=True,
            tile_size=1024,
            kwargs_tifffile=dict(photometric="rgb", planarconfig="separate"),
        )
    return 0


def viz_coarse(r1, r2, level1, level2, channel1, channel2, mx):
    try:
        import napari
    except ImportError:
        return
    import dask.array as da

    v = napari.Viewer()
    is_bf1 = palom.img_util.is_brightfield_img(r1.pyramid[-1][channel1])
    is_bf2 = palom.img_util.is_brightfield_img(r2.pyramid[-1][channel2])
    inv = {True: da.invert, False: da.array}
    v.add_image(
        [inv[is_bf1](p[channel1]) for p in r1.pyramid[level1:]],
        colormap="bop orange",
        blending="additive",
    )
    v.add_image(
        [inv[is_bf2](p[channel2]) for p in r2.pyramid[level2:]],
        affine=palom.img_util.to_napari_affine(mx),
        colormap="bop blue",
        blending="additive",
    )
    return v


def get_reader(path):
    path = pathlib.Path(path)
    if path.suffix in [".svs", ".ndpi"]:
        return palom.reader.SvsReader
    else:
        return palom.reader.OmePyramidReader


def set_matplotlib_font(font_size=12):
    font_families = matplotlib.rcParams["font.sans-serif"]
    if font_families[0] != "Arial":
        font_families.insert(0, "Arial")
    matplotlib.rcParams["pdf.fonttype"] = 42
    matplotlib.rcParams.update({"font.size": font_size})


def save_all_figs(dpi=300, format="pdf", out_dir=None, prefix=None):
    figs = [plt.figure(i) for i in plt.get_fignums()]
    if prefix is not None:
        for f in figs:
            if f._suptitle:
                f.suptitle(f"{prefix} {f._suptitle.get_text()}")
            else:
                f.suptitle(prefix)
    names = [f._suptitle.get_text() if f._suptitle else "" for f in figs]
    out_dir = pathlib.Path(out_dir)
    out_dir.mkdir(exist_ok=True, parents=True)

    for f, n, nm in zip(figs, plt.get_fignums(), names):
        f.savefig(out_dir / f"{n}-{nm}.{format}", dpi=dpi, bbox_inches="tight")
        plt.close(f)


def set_subplot_size(w, h, ax=None):
    """w, h: width, height in inches"""
    if not ax:
        ax = plt.gca()
    l = ax.figure.subplotpars.left
    r = ax.figure.subplotpars.right
    t = ax.figure.subplotpars.top
    b = ax.figure.subplotpars.bottom
    figw = float(w) / (r - l)
    figh = float(h) / (t - b)
    ax.figure.set_size_inches(figw, figh)


if __name__ == "__main__":
    import fire
    import sys

    fire.Fire(align_he)

    if ("--viz_coarse_napari" in sys.argv) or ("-v" in sys.argv):
        try:
            import napari
        except ImportError:
            print("napari is not installed")
        else:
            napari.run()

    """
    Example 1: inspect coarse alignment using napari
    python align_he.py \
        Z:\RareCyte-S3\P54_CRCstudy_Bridge\P54_S33_Full_Or6_A31_C90c_HMS@20221025_001610_632297.ome.tiff \
        "X:\crc-scans\histowiz scans\20230105-orion_2_cycles\22199$P54_33_HE$US$SCAN$OR$001 _104050.svs" \
        "X:\crc-scans\histowiz scans\20230105-orion_2_cycles\test" \
        --px_size1 0.325 --only_qc --only_coarse --viz_coarse_napari

    Example 2: process pair and output registered image
    python align_he.py \
        Z:\RareCyte-S3\P54_CRCstudy_Bridge\P54_S33_Full_Or6_A31_C90c_HMS@20221025_001610_632297.ome.tiff \
        "X:\crc-scans\histowiz scans\20230105-orion_2_cycles\22199$P54_33_HE$US$SCAN$OR$001 _104050.svs" \
        "X:\crc-scans\histowiz scans\20230105-orion_2_cycles\test" \
        --px_size1 0.325
   
    """



import numpy as np
import dask.array as da
import skimage.exposure
import sklearn.linear_model
from loguru import logger
import tqdm.dask

from . import register
from . import block_affine
from . import img_util


def block_affine_transformed_moving_img(ref_img, moving_img, mxs, is_mask=False):
    assert img_util.is_single_channel(ref_img)
    return_slice = slice(None)
    if img_util.is_single_channel(moving_img) and moving_img.ndim == 2:
        # return 2d image if ndim of moving_img is 2
        moving_img = moving_img[np.newaxis]
        return_slice = 0
    return da.array([
        da.map_blocks(
            block_affine.block_affine_dask,
            mxs,
            src_array=c,
            chunks=ref_img.chunks,
            dtype=moving_img.dtype,
            is_mask=is_mask
        )
        for c in moving_img
    ])[return_slice]


def _pc(img1, img2, mask, **pcc_kwargs):
    if not np.all(mask):
        return (np.inf, np.inf), np.inf
    return register.phase_cross_correlation(img1, img2, **pcc_kwargs)


def block_shifts(ref_img, moving_img, mask=True, pcc_kwargs=None):
    default_pcc_kwargs = dict(sigma=0, upsample=1)
    if pcc_kwargs is None:
        pcc_kwargs = {}
    return da.map_blocks(
        lambda a, b, m: np.atleast_2d(
            _pc(a, b, m, **{**default_pcc_kwargs, **pcc_kwargs})[0]
        ),
        ref_img,
        moving_img,
        mask,
        dtype=np.float32
    )


def constrain_block_shifts(shifts, grid_shape):
    distances = np.linalg.norm(shifts, axis=1)
    is_finite = np.isfinite(distances)
    # exclude np.inf when computing threshold
    threshold_distance = skimage.filters.threshold_triangle(
        distances[is_finite]
    )

    high_confidence_blocks = distances < threshold_distance

    lr = sklearn.linear_model.LinearRegression()
    block_coords = np.indices(grid_shape).reshape(2, -1).T
    lr.fit(
        block_coords[high_confidence_blocks],
        shifts[high_confidence_blocks]
    )
    predicted_shifts = lr.predict(block_coords)
    diffs = shifts - predicted_shifts
    distance_diffs = np.linalg.norm(diffs, axis=1)
    passed = (
        distance_diffs <
        # exclude np.inf when computing threshold
        skimage.filters.threshold_triangle(distance_diffs[is_finite])
    )
    fitted_shifts = shifts.copy()
    fitted_shifts[~passed] = predicted_shifts[~passed]
    return fitted_shifts


def viz_shifts(shifts, grid_shape, dcenter=None, ax=None):
    import matplotlib.pyplot as plt
    import matplotlib.colors
    distances = np.linalg.norm(shifts, axis=1)
    is_finite = np.isfinite(distances)
    if dcenter is None:
        # exclude np.inf when computing threshold
        dcenter = skimage.filters.threshold_triangle(distances[is_finite])
    # exclude np.inf when computing threshold
    dmin, dmax = np.percentile(distances[is_finite], (0, 100))
    divnorm = matplotlib.colors.TwoSlopeNorm(dcenter, dmin, dmax)
    colorbar_ticks = np.concatenate(
        [np.linspace(dmin, dcenter, 5), np.linspace(dcenter, dmax, 5)[1:]]
    )
    if ax is None:
        _, ax = plt.subplots()
    custom_cmap = matplotlib.colors.LinearSegmentedColormap.from_list(
        'cm_shift', np.vstack([
            plt.cm.plasma(np.linspace(0, 1, 256)),
            plt.cm.gray(np.linspace(0.5, 1, 256))
        ])
    )
    im = ax.imshow(distances.reshape(grid_shape), norm=divnorm, cmap=custom_cmap)
    cax = ax.inset_axes([1.04, 0.0, 0.02, 1])
    colorbar = plt.colorbar(im, cax=cax)
    colorbar.set_ticks(colorbar_ticks)
   
    return ax


def block_affine_matrices(mx, shifts):
       
    def shift_affine_mx(mx, shift):
        y, x = shift
        mx_shift = np.eye(3)
        mx_shift[:2, 2] = x, y
        return mx_shift @ mx

    mxs = [
        shift_affine_mx(mx, s)
        for s in shifts
    ]
    return np.array(mxs)


def block_affine_matrices_da(mxs, grid_shape):
    num_rows, num_cols = grid_shape
    grid = np.arange(num_rows * num_cols).reshape(grid_shape)
    mxs = np.vstack([
        np.hstack(mxs[r])
        for r in grid
    ])
    return da.from_array(mxs, chunks=3)


class Aligner:

    def __init__(
        self,
        ref_img,
        moving_img,
        ref_thumbnail,
        moving_thumbnail,
        ref_thumbnail_down_factor,
        moving_thumbnail_down_factor
    ) -> None:
        self.ref_img=ref_img
        self.moving_img=moving_img
        self.ref_thumbnail=ref_thumbnail
        self.moving_thumbnail=moving_thumbnail
        self.ref_thumbnail_down_factor=ref_thumbnail_down_factor
        self.moving_thumbnail_down_factor=moving_thumbnail_down_factor

    def coarse_register_affine(self, **kwargs):
        default_kwargs = {
            'n_keypoints': 2000,
            'plot_match_result': True
        }
        default_kwargs.update(kwargs)
        ref_img = self.ref_thumbnail
        moving_img = self.moving_thumbnail
        affine_matrix = register.feature_based_registration(
            ref_img, moving_img,
            **default_kwargs
        )
        self.coarse_affine_matrix = np.vstack(
            [affine_matrix, [0, 0, 1]]
        )
   
    @property
    def affine_matrix(self):
        if not hasattr(self, 'coarse_affine_matrix'):
            self.coarse_register_affine()
        affine = skimage.transform.AffineTransform
        mx_ref = affine(scale=1/self.ref_thumbnail_down_factor).params
        mx_moving = affine(scale=1/self.moving_thumbnail_down_factor).params
        affine_matrix = (
            np.linalg.inv(mx_ref) @
            self.coarse_affine_matrix.copy() @
            mx_moving
        )
        return affine_matrix
   
    @property
    def tform(self):
        return skimage.transform.AffineTransform(
            matrix=self.affine_matrix
        )
   
    def affine_transformed_moving_img(self, mxs=None):
        if mxs is None:
            mxs = self.affine_matrix
        ref_img = self.ref_img
        moving_img = self.moving_img

        return block_affine_transformed_moving_img(
            ref_img, moving_img, mxs
        )
   
    def compute_shifts(self, mask=True, pcc_kwargs=None):
        logger.info(f"Computing block-wise shifts")
        ref_img = self.ref_img
        moving_img = self.affine_transformed_moving_img(self.affine_matrix)
        shifts_da = block_shifts(ref_img, moving_img, mask, pcc_kwargs=pcc_kwargs)
        with tqdm.dask.TqdmCallback(
            ascii=True, desc='Computing shifts',
        ):
            shifts = shifts_da.compute()
        self.shifts = shifts.reshape(-1, 2)

    @property
    def grid_shape(self):
        return self.ref_img.numblocks

    @property
    def num_blocks(self):
        return self.ref_img.npartitions

    def constrain_shifts(self):
        if not hasattr(self, 'original_shifts'):
            self.original_shifts = self.shifts.copy()
        self.shifts = constrain_block_shifts(
            self.original_shifts,
            self.grid_shape
        )
   
    @property
    def block_affine_matrices(self):
        mx = self.affine_matrix
        shifts = self.shifts
        return block_affine_matrices(mx, shifts)

    @property
    def block_affine_matrices_da(self):
        return block_affine_matrices_da(
            self.block_affine_matrices,
            self.grid_shape
        )

    def overlay_grid(self, ax=None):
        import matplotlib.pyplot as plt
        img = self.ref_thumbnail
        img = skimage.exposure.rescale_intensity(img, out_range=np.uint16)
        shape = self.grid_shape
        grid = np.arange(np.multiply(*shape)).reshape(shape)
        h, w = np.divide(
            img.shape,
            np.divide(self.ref_img.chunksize, self.ref_thumbnail_down_factor)
        )
        cmap = 'gray_r' if img_util.is_brightfield_img(img) else 'gray'
        func = np.array if img_util.is_brightfield_img(img) else np.log1p

        if ax is None:
            _, ax = plt.subplots()
        ax.imshow(
            func(img),
            cmap=cmap,
            extent=(-0.5, w-0.5, h-0.5, -0.5)
        )
        # checkerboard pattern
        checkerboard = np.indices(shape).sum(axis=0) % 2
        if hasattr(self, 'shifts'):
            shifts = getattr(self, 'original_shifts', self.shifts)
            checkerboard = checkerboard.astype(float)
            checkerboard.flat[~np.all(np.isfinite(shifts), axis=1)] = np.nan
        ax.imshow(checkerboard, cmap='cool', alpha=0.2)
        return grid
   
    def plot_shifts(self):
        import matplotlib.pyplot as plt
        fig, axs = plt.subplots(1, 2, sharex=True, sharey=True)
        self.overlay_grid(axs[0])
        shifts = getattr(self, 'original_shifts', self.shifts)
        viz_shifts(shifts, self.grid_shape, ax=axs[1])
        return fig


def get_aligner(
    reader1, reader2,
    level1=0, level2=0,
    channel1=0, channel2=0,
    thumbnail_level1=-1, thumbnail_level2=-1,
    thumbnail_channel1=None, thumbnail_channel2=None
):
    if None in [thumbnail_level1, thumbnail_level2]:
        thumbnail_level1, thumbnail_level2 = match_thumbnail_level(
            [reader1, reader2]
        )
    if thumbnail_level1 <= -1: thumbnail_level1 += len(reader1.pyramid)
    if thumbnail_level2 <= -1: thumbnail_level2 += len(reader2.pyramid)
    thumbnail_channel1 = thumbnail_channel1 or channel1
    thumbnail_channel2 = thumbnail_channel2 or channel2
    return Aligner(
        reader1.read_level_channels(level1, channel1), 
        reader2.read_level_channels(level2, channel2),
        reader1.read_level_channels(thumbnail_level1, thumbnail_channel1),
        reader2.read_level_channels(thumbnail_level2, thumbnail_channel2),
        reader1.level_downsamples[thumbnail_level1] / reader1.level_downsamples[level1],
        reader2.level_downsamples[thumbnail_level2] / reader2.level_downsamples[level2]
    )


def match_thumbnail_level(readers):
    assert len(readers) > 1
    level_px_sizes = [
        {
            rr.pixel_size*vv: kk
            for kk, vv in rr.level_downsamples.items()
        }
        for rr in readers
    ]
    px_sizes = [sorted(ss.keys()) for ss in level_px_sizes]
    target_px_size = min([max(ss) for ss in px_sizes])
    target_levels = [
        lps[ps[np.argmin(np.abs(np.array(ps) - target_px_size))]]
        for ps, lps in zip(px_sizes, level_px_sizes)
    ]
    return target_levels


import cv2
import matplotlib.pyplot as plt
import numpy as np
import skimage.exposure
import skimage.filters
import skimage.transform

from . import img_util


IS_BF_IMG = img_util.is_brightfield_img


def make_img_pairs(img1, img2, auto_invert_intensity=True, auto_mask=False):
    img1 = np.asarray(img1).astype(np.float32)
    img2 = np.asarray(img2).astype(np.float32)
    compare_funcs = [
        np.less if IS_BF_IMG(i) else np.greater
        for i in (img1, img2)
    ]
    if not auto_invert_intensity:
        compare_funcs = [compare_funcs[0]]*2
    imgs_otsu = [
        f(i, skimage.filters.threshold_otsu(i)).astype(np.uint8)
        for (i, f) in zip((img1, img2), compare_funcs)
    ]
    imgs_tri = [
        f(i, skimage.filters.threshold_triangle(i)).astype(np.uint8)
        for (i, f) in zip((img1, img2), compare_funcs)
    ]
    if auto_invert_intensity:
        img1, img2 = match_bf_fl_histogram(img1, img2, auto_mask)
    else:
        match_func = skimage.exposure.match_histograms
        if auto_mask:
            match_func = masked_match_histograms
        img2 = match_func(img2, img1)

    imgs_whiten = [
        img_util.whiten(i, 1)
        for i in (img1, img2)
    ]
    return [
        imgs_otsu,
        imgs_tri,
        (img1, img2),
        imgs_whiten
    ]


def match_bf_fl_histogram(img1, img2, auto_mask=False):
    img1 = img1.astype(np.float32)
    img2 = img2.astype(np.float32)
    # TODO does it make a difference to min/max rescale before histogram
    # matching?
    is_bf_img1, is_bf_img2 = [
        IS_BF_IMG(i)
        for i in (img1, img2)
    ]
    match_func = skimage.exposure.match_histograms
    if auto_mask:
        match_func = masked_match_histograms
    if is_bf_img1 == is_bf_img2:
        return img1, match_func(img2, img1)
    elif is_bf_img1:
        return img1, match_func(-img2, img1)
    elif is_bf_img2:
        return match_func(-img1, img2), img2


def masked_match_histograms(img, ref_img):
    if IS_BF_IMG(img) != IS_BF_IMG(ref_img):
        bg = ['dark', 'light']
        print('`img` and `ref_img` detected as different types:')
        print(f"    `img` detected as {bg[IS_BF_IMG(img)]}-background image")
        print(f"    `ref_img` detected as {bg[IS_BF_IMG(ref_img)]}-background image")

    # downsize images to ~1000 px for speed
    shape_max = max(*img.shape, *ref_img.shape)
    downsize_factor = int(np.floor(shape_max / 500))
    if downsize_factor < 1:
        downsize_factor = 1
    mask = img_util.entropy_mask(
        img_util.cv2_downscale_local_mean(img, downsize_factor)
    )
    ref_mask = img_util.entropy_mask(
        img_util.cv2_downscale_local_mean(ref_img, downsize_factor)
    )
    repeats = (downsize_factor, downsize_factor)
    shape = img.shape
    ref_shape = ref_img.shape
    mask = img_util.repeat_2d(mask, repeats)[:shape[0], :shape[1]]
    ref_mask = img_util.repeat_2d(ref_mask, repeats)[:ref_shape[0], :ref_shape[1]]
    matched_img = np.zeros_like(img)
    # NOTE this does not handle inverted matching, both image must be the same
    # type. E.g. dark background, light signal
    matched_img[mask] = skimage.exposure.histogram_matching._match_cumulative_cdf(
        img[mask], ref_img[ref_mask]
    )
    matched_img[~mask] = ref_img[~ref_mask].mean()
    # matched_img[~mask] = skimage.exposure.histogram_matching._match_cumulative_cdf(
    #     img[~mask], ref_img[~ref_mask]
    # )
    return matched_img


def plot_img_keypoints(imgs, keypoints):
    fig, axs = plt.subplots(1, len(imgs))
    for i, k, a in zip(imgs, keypoints, axs):
        a.imshow(cv2.drawKeypoints(
            i, k, None,
            flags=cv2.DRAW_MATCHES_FLAGS_DEFAULT
        ))
        a.set_title(len(k))
    return


def get_flip_mx(img_shape, flip_axis):
    assert flip_axis in [0, 1, (0, 1), (1, 0)]
    mx = np.eye(3)
    offset_xy = np.array(img_shape)[::-1] - 1
    if type(flip_axis) == int:
        index = int(not flip_axis)
        mx[index, index] = -1
        mx[index, 2] = offset_xy[index]
        return mx
    mx[:2, :2] *= -1
    mx[:2, 2] = offset_xy
    return mx


def get_rot90_mx(img_shape, k):
    assert k in range(4)
    degree = -k*90
    h, w = img_shape
    translation = {
        0: (0, 0),
        1: (0, w-1),
        2: (w-1, h-1),
        3: (h-1, 0)
    }
    return skimage.transform.AffineTransform(
        rotation=np.deg2rad(degree),
        translation=translation[k]
    ).params



import functools
import inspect
import itertools
import warnings

import cv2
import matplotlib.pyplot as plt
import numpy as np
import scipy.fft
import skimage.exposure
import skimage.feature
import skimage.registration
from loguru import logger

from . import img_util, register_util

if hasattr(skimage.registration, 'phase_cross_correlation'):
    register_translation = skimage.registration.phase_cross_correlation
else:
    register_translation = skimage.feature.register_translation


#
# Image-based registration
#
def phase_cross_correlation(img1, img2, sigma, upsample=10, module='skimage'):
    assert module in ['cv2', 'skimage']

    if (np.unique(img1).size == 1) | (np.unique(img2).size == 1):
        # FIXME does this make sense?
        return (np.inf, np.inf), np.inf
    
    img1w = img_util.whiten(img1, sigma)
    img2w = img_util.whiten(img2, sigma)
   
    if module == 'skimage':
        with warnings.catch_warnings():
            warnings.filterwarnings(
                # two patterns observed
                # 1. invalid value encountered in true_divide
                # 2. invalid value encountered in divide
                'ignore', 'invalid value encountered in',
                RuntimeWarning,
            )
            kwargs = dict(upsample_factor=upsample)
            # `normalization` kwarg was introduced in skimage v0.19
            if 'normalization' in inspect.signature(register_translation).parameters:
                kwargs.update(normalization=None)
            shift, _error, _phasediff = register_translation(
                img1w, img2w, **kwargs
            )
   
    elif module == 'cv2':
        shift_xy, _response = cv2.phaseCorrelate(img1w, img2w)
        shift = shift_xy[::-1]

    # At this point we may have a shift in the wrong quadrant since the FFT
    # assumes the signal is periodic. We test all four possibilities and return
    # the shift that gives the highest direct correlation (sum of products).
    shape = np.array(img1.shape)
    shift_pos = (shift + shape) % shape
    shift_neg = shift_pos - shape
    shifts = list(itertools.product(*zip(shift_pos, shift_neg)))
    correlations = [
        np.abs(np.sum(img1w * cv2_translate(img2w, s)))
        for s in shifts
    ]
    idx = np.argmax(correlations)
    shift = shifts[idx]
    correlation = correlations[idx]
    total_amplitude = np.linalg.norm(img1w) * np.linalg.norm(img2w)
    if correlation > 0 and total_amplitude > 0:
        error = -np.log(correlation / total_amplitude)
    else:
        error = np.inf
    return shift, error


def cv2_translate(img, shift):
    assert img.ndim == len(shift) == 2
    sy, sx = shift
    return cv2.warpAffine(
        img,   
        np.array([[1, 0, sx], [0, 1, sy]], dtype=float),
        img.shape[::-1]
    )


def normalized_phase_correlation(img1, img2, sigma):
    w1 = img_util.whiten(img1, sigma)
    w2 = img_util.whiten(img2, sigma)
    corr = scipy.fft.fftshift(np.abs(scipy.fft.ifft2(
        scipy.fft.fft2(w1) * scipy.fft.fft2(w2).conj()
    )))
    corr /= (np.linalg.norm(w1) * np.linalg.norm(w2))
    return corr


#
# Feature-based registration
#
def feature_based_registration(
    img_left, img_right,
    test_flip=False, test_intensity_invert=False,
    n_keypoints=1000, plot_match_result=False,
    plot_individual_result=False, ransacReprojThreshold=5,
    # apply entropy mask during histogram matching
    # turn this off if the image is not a WSI
    auto_mask=False,
    kwarg_test_flip_and_intensity_invert=None
):
    _kwarg_test_flip_and_intensity_invert = {'auto_mask': auto_mask}
    if kwarg_test_flip_and_intensity_invert is None:
        kwarg_test_flip_and_intensity_invert = {}

    if test_flip == test_intensity_invert == True:
        kwarg_test_flip_and_intensity_invert['test_only'] = None
    if test_flip != test_intensity_invert:
        test_only = 'flip'
        if test_intensity_invert:
            test_only = 'intensity_invert'
        kwarg_test_flip_and_intensity_invert['test_only'] = test_only
    
    flip_func = np.array
    intensity_scalar = 1
    auto_invert_intensity = True
    mx_fr = np.eye(3)

    if test_flip | test_intensity_invert:
        test_result = test_flip_and_intensity_invert(
            img_left, img_right,
            **{
                **_kwarg_test_flip_and_intensity_invert,
                **kwarg_test_flip_and_intensity_invert
            }
        )
        flip_func = test_result['flip_func']
        intensity_scalar = test_result['intensity_scalar']
        if test_intensity_invert:
            auto_invert_intensity = False
        if flip_func != np.array:
            mx_fr = register_util.get_flip_mx(img_right.shape, 0)

    img_right = flip_func(img_right * intensity_scalar)
    mx_affine = ensambled_match(
        img_left, img_right,
        n_keypoints, plot_match_result,
        plot_individual_result, ransacReprojThreshold,
        auto_invert_intensity=auto_invert_intensity,
        auto_mask=auto_mask,
    )
    if mx_affine is None: 
        logger.warning('Feature matching failed. Returning identity matrix as placeholder') 
        mx_affine = np.eye(3)[:2] 
    mx_affine = (np.vstack([mx_affine, [0, 0, 1]]) @ mx_fr)[:2, :]
    return mx_affine


def test_flip_and_intensity_invert(
        img_left, img_right,
        max_size=500,
        auto_mask=False,
        test_only=None
    ):
    assert test_only in [None, 'flip', 'intensity_invert']
    flip_funcs = [np.array, np.flipud]
    intensity_scalars = [1, -1]
    
    if test_only == 'flip':
        intensity_scalars.pop(1)
    
    if test_only == 'intensity_invert':
        flip_funcs.pop(1)
    
    # downsize images to ~500 px for speed
    shape_max = max(*img_left.shape, *img_right.shape)
    downsize_factor = int(np.floor(shape_max / max_size))
    if downsize_factor > 1:
        img_left = img_util.cv2_downscale_local_mean(img_left, downsize_factor)
        img_right = img_util.cv2_downscale_local_mean(img_right, downsize_factor)
    
    match_results = []
    for flip_func, scalar in itertools.product(
            flip_funcs, intensity_scalars
    ):
        n_matches = ensambled_match(
            img_left,
            flip_func(scalar * img_right),
            return_match_mask=True,
            auto_invert_intensity=False,
            auto_mask=auto_mask
        )[1].sum()
        match_results.append((n_matches, flip_func, scalar))
    for rr in match_results: print(rr)
    best = sorted(match_results, key=lambda x: x[0], reverse=True)[0]
    return {'flip_func': best[1], 'intensity_scalar': best[2]}


def ensambled_match(
    img_left, img_right,
    n_keypoints=1000, plot_match_result=False,
    plot_individual_result=False, ransacReprojThreshold=5,
    return_match_mask=False,
    # try to infer imaging modalities from the images and perform intensity
    # inversion when needed; turn this off if the imaging modalities are known
    # to be the same
    auto_invert_intensity=True,
    # apply entropy mask during histogram matching
    # turn this off if the image is not a WSI
    auto_mask=False
):
    img_pairs = register_util.make_img_pairs(
        img_left, img_right,
        auto_invert_intensity=auto_invert_intensity,
        auto_mask=auto_mask
    )
    img_left, img_right = img_pairs[2]

    all_found = [
        cv2_feature_detect_and_match(
            *img_pair, n_keypoints=n_keypoints,
            plot_match_result=plot_individual_result
        )
        for img_pair in img_pairs
    ]
    all_src = np.vstack([i[0] for i in all_found])
    all_dst = np.vstack([i[1] for i in all_found])

    t_matrix, mask = cv2.estimateAffine2D(
        all_dst, all_src,
        method=cv2.RANSAC,
        ransacReprojThreshold=ransacReprojThreshold,
        maxIters=5000
    )
    if plot_match_result == True:
        _, ax = plt.subplots()
        skimage.feature.plot_matches(
            ax,
            np.log(skimage.exposure.rescale_intensity(img_left, out_range=(1, 4095))),
            np.log(skimage.exposure.rescale_intensity(img_right, out_range=(1, 4095))),
            np.fliplr(all_src), np.fliplr(all_dst),
            np.arange(len(all_src)).repeat(2).reshape(-1, 2)[mask.flatten()>0],
            keypoints_color=np.divide([255, 215, 0, 50], 255),
            matches_color='deepskyblue',
            only_matches=False
        )
        for line in ax.get_lines():
            line.set_alpha(0.5)
            line.set_linewidth(.5)

    return (t_matrix, mask) if return_match_mask else t_matrix


def cv2_feature_detect_and_match(
    img_left, img_right, n_keypoints=1000,
    plot_keypoint_result=False, plot_match_result=False
):
    img_left, img_right = [
        img_util.cv2_to_uint8(i)
        for i in (img_left, img_right)
    ]
    descriptor_extractor = cv2.ORB_create(n_keypoints, edgeThreshold=0)

    keypoints_left, descriptors_left = descriptor_extractor.detectAndCompute(
        np.dstack(3*(img_left,)), None
    )
    keypoints_right, descriptors_right = descriptor_extractor.detectAndCompute(
        np.dstack(3*(img_right,)), None
    )
    if plot_keypoint_result == True:
        register_util.plot_img_keypoints(
            [img_left, img_right], [keypoints_left, keypoints_right]
        )
    logger.debug(f"keypts L:{len(keypoints_left)}, keypts R:{len(keypoints_right)}")
    if len(keypoints_left) == 0 or len(keypoints_right) == 0:
        return np.empty((1, 2)), np.empty((1, 2))

    bf_matcher = cv2.BFMatcher(cv2.NORM_HAMMING, crossCheck=True)
    matches = bf_matcher.match(descriptors_left, descriptors_right)

    src_pts = np.float32(
        [keypoints_left[m.queryIdx].pt for m in matches]
    )
    dst_pts = np.float32(
        [keypoints_right[m.trainIdx].pt for m in matches]
    )
    t_matrix, mask = cv2.estimateAffine2D(
        dst_pts, src_pts,
        method=cv2.RANSAC, ransacReprojThreshold=30, maxIters=5000
    )
    if plot_match_result == True:
        plt.figure()
        imgmatch_ransac = cv2.drawMatches(
            img_left, keypoints_left,
            img_right, keypoints_right,
            matches, None,
            matchColor=(0, 255, 0), singlePointColor=None,
            matchesMask=mask.flatten(),
            flags=cv2.DRAW_MATCHES_FLAGS_DEFAULT
        )
        plt.gca().imshow(imgmatch_ransac)
    return src_pts[mask.flatten()>0], dst_pts[mask.flatten()>0]


def match_test_flip_rotate(img_left, img_right, auto_mask=False):

    flip_funcs = [np.array] + [
        functools.partial(np.flip, axis=aa)
        for aa in (0, 1, (0, 1))
    ]
    rotate_funcs = [
        functools.partial(np.rot90, k=i)
        for i in range(4)
    ]
    flip_mxs = [np.eye(3)] + [
        register_util.get_flip_mx(img_right.shape, aa)
        for aa in (0, 1, (0, 1))
    ]
    rotate_mxs = [
        register_util.get_rot90_mx(img_right.shape, i)
        for i in range(4)
    ]

    # downsize images to ~500 px for speed
    shape_max = max(*img_left.shape, *img_right.shape)
    downsize_factor = int(np.floor(shape_max / 500))
    if downsize_factor > 1:
        img_left = img_util.cv2_downscale_local_mean(img_left, downsize_factor)
        img_right = img_util.cv2_downscale_local_mean(img_right, downsize_factor)

    n_matches = [
        ensambled_match(
            img_left, rr(ff(img_right)), return_match_mask=True, auto_mask=auto_mask
        )[1].sum()
        # only need half of the 4x4 combinations
        for ff, rr in itertools.product(flip_funcs[:2], rotate_funcs)
    ]
    best_flip, best_rotate = np.unravel_index(
        np.argmax(n_matches), (2, 4)
    )
    print(np.array(n_matches, int).reshape(2, 4))
    print(best_flip, best_rotate)

    # construct best flip and rotate func
    ff, rr = flip_funcs[best_flip], rotate_funcs[best_rotate]
    def flip_rotate_func(target_img):
        return rr(ff(target_img))

    return flip_rotate_func, rotate_mxs[best_rotate] @ flip_mxs[best_flip]



import numpy as np

# from . import register
import palom.register as register


def pc(img1, img2, sigma, mask=None):
    if mask is not None:
        if not mask:
            return np.array([0] * 3, dtype=np.float32).reshape(3, 1, 1)
    (r_shift, c_shift), error = register.phase_cross_correlation(
        img1, img2, sigma, 10
    )
    # NOTE invert shift direction from the registration func
    return np.array(
        [-r_shift, -c_shift, error], dtype=np.float32
    ).reshape(3, 1, 1)


def shifts_to_lab(shifts, max_radius=None, l_factor=50, ab_factor=100):
    if max_radius is not None:
        assert max_radius > 0
    # shifts is [np.inf, np.inf, np.inf] for blank block
    valid = np.all(np.isfinite(shifts), axis=0)
    # FIXME workaround to exclude masked blocks which are [0, 0, 0]
    valid &= ~np.all(shifts == 0, axis=0)
    
    distances = np.where(valid, np.linalg.norm(shifts[:2], axis=0), np.nan)
    lab = np.zeros((3, *shifts[0].shape))
    
    if max_radius is None:
        max_radius = np.nanmax(distances)
    if max_radius == 0:
        return lab
    
    # normalize the shift vectors
    distances = np.clip(distances, max_radius, None)
    lab[1:3, valid] = (shifts[:2, valid] / distances[valid])[::-1]
    lab[0] = np.linalg.norm(lab[1:3], axis=0)
    lab *= np.array([l_factor, ab_factor, ab_factor]).reshape(3, 1, 1)
    return lab


import dask.array as da
import dask.diagnostics
import palom


def optical_flow(
    p1, p2, ch1, ch2,
    block_size,
    mask=None,
    num_workers=4,
    chunk_size=4096
):
    # Use process parallelization
    chunk_size = int(np.ceil(chunk_size / block_size) * block_size)
    Reader = palom.reader.OmePyramidReader
    shape = Reader(p1).pyramid[0].shape[1:]
    ref = da.ones((1, *shape), chunks=chunk_size)

    mask_shape = da.ones(shape, chunks=block_size).numblocks
    if mask is None:
        mask = np.ones(mask_shape, dtype=bool)
    assert mask.shape == mask_shape
   
    def func(
        p1, p2, ch1, ch2, mask,
        block_info=None
    ):
        _, rloc, cloc = block_info[None]['array-location']
        chunk1 = Reader(p1).pyramid[0][ch1, slice(*rloc), slice(*cloc)]
        chunk2 = Reader(p2).pyramid[0][ch2, slice(*rloc), slice(*cloc)]
        chunk_mask = mask[
            slice(*np.ceil(np.divide(rloc, block_size)).astype(int)),
            slice(*np.ceil(np.divide(cloc, block_size)).astype(int))
        ]
        return block_optical_flow(chunk1, chunk2, block_size, mask=chunk_mask)
   
    shifts = da.map_blocks(
        func,
        p1=p1, p2=p2, ch1=ch1, ch2=ch2, mask=mask,
        dtype=float,
        chunks=ref.chunks,
    )

    with dask.diagnostics.ProgressBar():
        return shifts.compute(
            scheduler='processes', num_workers=num_workers
        )


import itertools
import tqdm.contrib.itertools
import skimage.util
import skimage.transform
def block_optical_flow(
    img1, img2, block_size,
    sigma=0, mask=None,
    show_progress=False
):
    img1 = np.asarray(img1)
    img2 = np.asarray(img2)
    block_shape = (block_size, block_size)
    block_shape = np.min(np.vstack([img1.shape, block_shape]), axis=0)
    wv_img1 = skimage.util.view_as_windows(img1, block_shape, block_shape)
    wv_img2 = skimage.util.view_as_windows(img2, block_shape, block_shape)
    h, w = wv_img1.shape[:2]
   
    wv_mask = np.ones((h, w, 1, 1), dtype=bool)
    if mask is not None:
        wv_mask = mask[..., np.newaxis, np.newaxis]
        wv_mask = wv_mask[..., :h, :w]
   
    out = np.zeros((3, h, w), dtype=np.float32)
    product_func = itertools.product
    if show_progress:
        product_func = tqdm.contrib.itertools.product
    idxs = product_func(range(h), range(w))
    for rr, cc in idxs:
        out[:, rr, cc] = pc(
            wv_img1[rr, cc], wv_img2[rr, cc], sigma, mask=wv_mask[rr, cc]
        ).ravel()
    return out


def reader_block_mask(reader, block_size, level=-1, channel=0):
    mask = palom.img_util.entropy_mask(
        reader.pyramid[level][channel]
    ).astype(float)
    target_shape = da.ones(
        reader.pyramid[0].shape[1:], chunks=block_size
    ).numblocks
    resized_mask = skimage.transform.resize(
        mask, target_shape, order=2
    )
    return resized_mask > 0.1


import skimage.exposure
def compose_thumbnail(p1, p2, ch1, ch2, log_intensity=True):
    Reader = palom.reader.OmePyramidReader
    r1 = Reader(p1)
    r2 = Reader(p2)
    func = np.log1p if log_intensity else np.array
    img1 = np.asarray(func(r1.pyramid[-1][ch1]))
    img2 = np.asarray(func(r2.pyramid[-1][ch2]))

    mask = palom.img_util.entropy_mask(img1)
    rimg1 = skimage.exposure.rescale_intensity(
        img1,
        in_range=(np.mean(img1[~mask]), np.percentile(img1[mask], 99.9)),
        out_range=np.float32
    )
    rimg2 = skimage.exposure.rescale_intensity(
        img2,
        in_range=(np.mean(img2[~mask]), np.percentile(img2[mask], 99.9)),
        out_range=np.float32
    )
    rimg2 = skimage.exposure.match_histograms(rimg2, rimg1)
    out = np.array([rimg1, rimg2, rimg1])
    return out


def plot_legend(
    shifts, max_radius, plot_scatter, plot_kde,
    pad_plot=False, plot_flow=False, ax=None
):

    import matplotlib.pylab as plt
    import scipy.ndimage

    # illuminant : {"A", "D50", "D55", "D65", "D75", "E"}
    if max_radius is None:
        max_radius = np.linalg.norm(shifts[:2], axis=0).max()
    lower, upper = -max_radius, max_radius
    if ax is None:
        _, ax = plt.subplots(1, 1)
    fig = ax.get_figure()
    legend_raster = np.mgrid[-1:1:501j, -1:1:501j]
    legend_raster[:, np.linalg.norm(legend_raster, axis=0) > 1] = 0

    ax.imshow(
        skimage.color.lab2rgb(
            np.dstack(shifts_to_lab(legend_raster, l_factor=50, ab_factor=100))
        ),
        # extent=(lower-.5, upper+.5, upper+.5, lower-.5)
        extent=(lower, upper, upper, lower)
    )
    xlim, ylim = ax.get_xlim(), ax.get_ylim()
    ax.set_xlim(xlim[0], xlim[1])
    ax.set_ylim(ylim[0], ylim[1])
    if pad_plot:
        ax.set_xlim(xlim[0]-.5, xlim[1]+.5)
        ax.set_ylim(ylim[0]+.5, ylim[1]-.5)
    if not (plot_scatter or plot_kde or plot_flow):
        return [fig]
    lab = shifts_to_lab(shifts, max_radius, l_factor=50, ab_factor=100)
    x_shifts, y_shifts = [
        # 100 is from `ab_factor`
        s[shifts[2] != 0].flatten() / 100 * max_radius
        for s in lab[1:3]
    ]
    if plot_scatter:
        ax.scatter(
            x_shifts, y_shifts,
            c=skimage.color.lab2rgb(
                np.dstack(shifts_to_lab(shifts, max_radius, l_factor=100, ab_factor=100)),
            ).reshape(-1, 3)[shifts[2].flatten() != 0],
            linewidths=0, s=4, alpha=.5
        )
    if plot_kde:
        # composite flow legend with kde
        density, ybins, xbins = np.histogram2d(
            y_shifts, x_shifts,
            bins=50, density=True,
        )
        density = scipy.ndimage.gaussian_filter(density, 1)
        ax.contour(
            density, levels=np.linspace(0, density.max(), 6)[1:],
            extent=(*xbins[[0, -1]], *ybins[[-1, 0]]),
            origin='upper',
            cmap='viridis',
            linewidths=0.5
        )
    if plot_flow:
        fig_flow, ax_flow = plt.subplots(1, 1)
        ax_flow.imshow(skimage.color.lab2rgb(np.dstack(lab)))
        return [fig, fig_flow]
    return [fig]


def get_img_extent(shape, downscale_factor):
    h, w = np.multiply(shape, downscale_factor)
    return (-0.5, w-0.5, h-0.5, -0.5)


import skimage.color
import matplotlib.pyplot as plt
import pathlib
from typing import List


def process_img_channel_pair(
    img_path: str | pathlib.Path,
    ref_channel: int,
    other_channels: List[int],
    block_size: int,
    max_radius=5,
    compute_mask:bool = True,
    num_workers: int = 4,
    as_script=True
):
    img_path = pathlib.Path(img_path)
    reader = palom.reader.OmePyramidReader(img_path)

    mask = None
    if compute_mask:
        mask = reader_block_mask(reader, block_size)
   
    num_cpus = dask.system.cpu_count()
    if num_workers > num_cpus:
        num_workers = num_cpus

    print()
    print(f"Processing {img_path.name} using {num_workers} cores")
    print(f"Block size: {block_size}; image shape {reader.pyramid[0].shape[1:]}")
    print("Reference channel:", ref_channel)
    all_shifts = []
    figures = []
    for channel in other_channels:
        print("    Processing channel:", channel)

        shifts = optical_flow(
            img_path, img_path, ref_channel, channel, block_size, mask,
            num_workers=num_workers
        )
        all_shifts.append(shifts)
        lab = shifts_to_lab(shifts, max_radius=max_radius)
        rgb = skimage.color.lab2rgb(lab, channel_axis=0)
        thumbnail = compose_thumbnail(
            img_path, img_path, ref_channel, channel, log_intensity=False
        )

        thumbnail_extent = get_img_extent(
            thumbnail[0].shape, reader.level_downsamples[len(reader.pyramid)-1]
        )
        flow_extent = get_img_extent(
            shifts[0].shape, block_size
        )

        fig = plt.figure(figsize=(12, 12))
        fig.suptitle(f"Optical flow: {img_path.name}")
        gs = fig.add_gridspec(1, 4, width_ratios=(1, 1, 1, 0.5))

        ax0 = fig.add_subplot(gs[0])
        ax0.imshow(np.dstack(thumbnail), extent=thumbnail_extent)
        ax0.set_title(f"Channel {ref_channel} (pink) and {channel} (green)")

        ax1 = fig.add_subplot(gs[1], sharex=ax0, sharey=ax0)
        ax1.imshow(thumbnail[1], extent=thumbnail_extent, alpha=0.5, cmap='gray')
        ax1.imshow(np.dstack(rgb), extent=flow_extent, alpha=0.5)
        ax1.label_outer()
        ax1.set_title(f"Overlay on channel {channel}")

        ax2 = fig.add_subplot(gs[2], sharex=ax0, sharey=ax0)
        ax2.imshow(np.dstack(rgb), extent=flow_extent)
        ax2.label_outer()
        ax2.set_title(f"Flow field")

        _ = plot_legend(
            shifts, max_radius, True, True, plot_flow=False, ax=fig.add_subplot(gs[3])
        )

        for ax in fig.get_axes():
            ax.set_anchor('N')
       
        plt.tight_layout()
        out_dir = img_path.parent / 'qc'
        out_dir.mkdir(exist_ok=True)
        fig.savefig(
            out_dir / f"flow-field-{img_path.stem}-channel-{ref_channel}-{channel}.png",
            bbox_inches='tight', dpi=144
        )
        figures.append(fig)
   
    if as_script: return
    return all_shifts, figures


if __name__ == '__main__':
    import fire
    fire.Fire(process_img_channel_pair)

    """
    NAME
        flow.py

    SYNOPSIS
        flow.py IMG_PATH REF_CHANNEL OTHER_CHANNELS BLOCK_SIZE <flags>

    POSITIONAL ARGUMENTS
        IMG_PATH
            Type: str | pathlib.Path
        REF_CHANNEL
            Type: int
        OTHER_CHANNELS
            Type: List
        BLOCK_SIZE
            Type: int

    FLAGS
        -m, --max_radius=MAX_RADIUS
            Default: 5
        -c, --compute_mask=COMPUTE_MASK
            Type: bool Default: True
        -n, --num_workers=NUM_WORKERS
            Type: int Default: 4
        -a, --as_script=AS_SCRIPT
            Default: True

    ---
   
    NOTE: OTHER_CHANNELS is a list of integers, use [1] to indicate the
    second channel and [1,2,3] (do not include space) for second, third and
    fourth channels.

    Ref https://google.github.io/python-fire/guide/#argument-parsing
   
    Examples

    python flow.py \
        "Z:\RareCyte-S3\P54_CRCstudy_Bridge\S32-Tonsil-P54_Strip_P76.ome.tif" \
        0 [1,2,3] \
        128

    python flow.py \
        "Z:\RareCyte-S3\P54_CRCstudy_Bridge\S32-Tonsil-P54_Strip_P76.ome.tif" \
        0 1,2,3 \
        128

    python flow.py \
        "Z:\RareCyte-S3\P54_CRCstudy_Bridge\S32-Tonsil-P54_Strip_P76.ome.tif" \
        0 [1] \
        128
    """


import numpy as np
import dask.array as da
import cv2
import skimage.transform
import skimage.filters
import skimage.morphology

def cv2_pyramid(img, max_size=1024):
    size_max = np.array(img.shape).max()
    pyramid = [img]
    while size_max > max_size:
        img = cv2.pyrDown(img)
        pyramid.append(img)
        size_max //= 2
    return pyramid


def whiten(img, sigma=1):
    border_mode = cv2.BORDER_REFLECT
    g_img = img if sigma == 0 else cv2.GaussianBlur(
        img, (0, 0), sigma,
        borderType=border_mode
    ).astype(np.float32)
    log_img = cv2.Laplacian(
        g_img, cv2.CV_32F, ksize=1,
        borderType=border_mode
    )
    # log_img = cv2.convertScaleAbs(log_img)
    return log_img


def cv2_to_uint8(img):
    if np.issubdtype(img.dtype, bool):
        img = img.astype(np.int8)
    return cv2.normalize(
        src=img, dst=None,
        alpha=255, beta=0,
        norm_type=cv2.NORM_MINMAX,
        dtype=cv2.CV_8U
    )


def is_single_channel(img):
    is_2d = img.ndim == 2
    is_flat_3d = img.ndim == 3 and img.shape[0] == 1
    return is_2d or is_flat_3d


def entropy_mask(img, kernel_size=9):
    img = skimage.exposure.rescale_intensity(
        img, out_range=np.uint8
    ).astype(np.uint8)
    entropy = skimage.filters.rank.entropy(img, np.ones((kernel_size, kernel_size)))
    return entropy > skimage.filters.threshold_otsu(entropy)


def is_brightfield_img(img, max_size=100):
    img = np.array(img)
    downscale_factor = int(max(img.shape) / max_size)
    if downscale_factor > 1:
        img = cv2_downscale_local_mean(img, downscale_factor)
    mask = entropy_mask(img)
    # is using mean better?
    return np.median(img[mask]) < np.median(img[~mask])


def repeat_2d(arr, repeats):
    assert arr.ndim == 2
    assert len(repeats) == 2
    r0, r1 = repeats
    return np.repeat(
        np.repeat(arr, r0, axis=0), r1, axis=1
    )


def block_labeled_mask(img_shape, block_shape, out_chunks=None):
    assert len(img_shape) == 2
    if len(block_shape) == 1:
        block_shape = (block_shape[0], block_shape[0])
    da_template = da.zeros(img_shape, chunks=block_shape)
    unit_mask = np.indices(da_template.numblocks).sum(axis=0) % 2
    unit_mask = skimage.morphology.label(
        unit_mask, connectivity=1, background=-1
    ).astype(np.int32)
    full_mask = np.repeat(
        np.repeat(unit_mask, block_shape[0], axis=0),
        block_shape[1],
        axis=1
    )[:img_shape[0], :img_shape[1]]
    if out_chunks is None:
        out_chunks = block_shape
    return da.from_array(full_mask, chunks=out_chunks)


def to_napari_affine(mx):
    ul = np.flip(mx[:2, :2], (0, 1))
    rows = np.hstack([ul, np.flipud(mx[:2, 2:3])])
    return np.vstack((rows, [0, 0, 1]))


# orders of magnitute faster than skimage.transform.downscale_local_mean
# also the gives sensible values of pixels on the edge
def cv2_downscale_local_mean(img, factor):
    assert img.ndim in [2, 3]
    img = np.asarray(img)
    axis_moved = False
    channel_ax = np.argmin(img.shape)
    if (img.ndim == 3) & (channel_ax != 2):
        img = np.moveaxis(img, channel_ax, 2)
        axis_moved = True
    simg = cv2.blur(
        img, ksize=(factor, factor), anchor=(0, 0)
    )[::factor, ::factor]
    if axis_moved:
        simg = np.moveaxis(simg, 2, channel_ax)
    return simg


from __future__ import annotations

import pathlib

import dask.array as da
import numpy as np
import ome_types
import pint
import tifffile
import zarr
from loguru import logger

from . import pyramid as pyramid_util


class DaPyramidChannelReader:

    def __init__(
        self,
        pyramid: list[da.Array],
        channel_axis: int
    ) -> None:
        self.pyramid = pyramid
        self.channel_axis = channel_axis
        if self.validate_pyramid(self.pyramid, self.channel_axis):
            self.pyramid = self.normalize_axis_order()
            self.pyramid = self.auto_format_pyramid(self.pyramid)

    @staticmethod
    def validate_pyramid(pyramid: list[da.Array], channel_axis:int) -> bool:
        for i, level in enumerate(pyramid):
            assert level.ndim == 3, ''
            if np.argmin(level.shape) != channel_axis:
                logger.warning(
                    f"level {i} has shape of {level.shape} while given"
                    f" `channel_axis` is {channel_axis}"
                )
        return True
   
    def normalize_axis_order(self):
        if self.channel_axis == 0:
            return self.pyramid
        return [
            da.moveaxis(level, self.channel_axis, 0)
            for level in self.pyramid
        ]

    def read_level_channels(
        self,
        level: int,
        channels: int | list[int]
    ) -> da.Array:
        target_level = self.pyramid[level]
        return target_level[channels]

    @staticmethod
    def auto_format_pyramid(
        pyramid: list[da.Array],
    ) -> list[da.Array]:
        first = pyramid[0]
        if len(pyramid) > 1: return pyramid
        # Assumption: if the image is pyramidal, it must also be tiled
        if max(first.shape) < 1024: return pyramid
        logger.warning(
                f'Unable to detect pyramid levels, it may take a while'
                f' to compute thumbnails during coarse alignment'
            )
        if first.numblocks[1:3] == (1, 1):
            first = first.rechunk((1, 1024, 1024))
        pyramid_setting = pyramid_util.PyramidSetting(downscale_factor=2)
        num_levels = pyramid_setting.num_levels(first.shape[1:3])
        return [
            da.coarsen(
                np.mean,
                first,
                {0:1, 1:2**i, 2:2**i},
                trim_excess=True
            ).astype(first.dtype)
            for i in range(num_levels)
        ]

    @property
    def level_downsamples(self) -> dict[int, int]:
        return {
            i: round(self.pyramid[0].shape[1] / level.shape[1])
            for i, level in enumerate(self.pyramid)
        }
   
    @property
    def pixel_dtype(self) -> np.dtype:
        return self.pyramid[0].dtype

    def get_thumbnail_level_of_size(self, size: float) -> int:
        shapes = [
            np.abs(np.mean(level.shape[1:3]) - size)
            for level in self.pyramid
        ]
        return np.argmin(shapes)


class OmePyramidReader(DaPyramidChannelReader):

    def __init__(
        self,
        path: str | pathlib.Path,
        pixel_size: float | None = None
    ) -> None:
        self.path = pathlib.Path(path)
        pyramid = self.pyramid_from_ometiff(self.path)
        channel_axis = 0
        self._pixel_size = pixel_size
        super().__init__(pyramid, channel_axis)

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['pyramid']
        state['path'] = state['path'].resolve()
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.__init__(path=state['path'], pixel_size=state['_pixel_size'])

    @staticmethod
    def pyramid_from_ometiff(path: str | pathlib.Path) -> list[da.Array]:
        with tifffile.TiffFile(path) as tif:
            num_series = len(tif.series)
            if num_series == 1:
                pyramid = tif.series[0].levels
            elif num_series > 1:
                pyramid = tif.series
            zarr_pyramid = [
                zarr.open(level.aszarr(), 'r')
                for level in pyramid
            ]
            da_pyramid = []
            for z in zarr_pyramid:
                if issubclass(type(z), zarr.hierarchy.Group):
                    da_level = da.from_zarr(z[0])
                else:
                    da_level = da.from_zarr(z)
                if da_level.ndim == 2:
                    da_level = da_level.reshape(1, *da_level.shape)
                if da_level.ndim == 3:
                    if da_level.shape[2] in (3, 4):
                        da_level = da.moveaxis(da_level, 2, 0)
                da_pyramid.append(da_level)
            return da_pyramid

    @property
    def pixel_size(self) -> float:
        if self._pixel_size is not None:
            return self._pixel_size
        try:
            # ome-types v0.4 does not have `parser` kwarg in `from_tiff`
            import inspect
            kwargs = dict(path=self.path, validate=False)
            keys = inspect.signature(ome_types.from_tiff).parameters
            if 'parser' in keys:
                kwargs.update(dict(parser='lxml'))
            ome = ome_types.from_tiff(**kwargs)
            px_size = ome.images[0].pixels.physical_size_x
            # convert length unit to µm
            unit = ome.images[0].pixels.physical_size_x_unit.value
            ureg = pint.UnitRegistry()
            px_size_micron = px_size * ureg(unit).to(ureg.micron).magnitude
            logger.info(
                f"Detected pixel size: {px_size_micron:.4f} µm"
            )
            self._pixel_size = px_size_micron
            return self._pixel_size
        except Exception:
            logger.warning(
                f'Unable to parse pixel size from {self.path.name};'
                f' assuming 1 µm. Use `_pixel_size` to set it manually'
            )
            self._pixel_size = 1
            return self._pixel_size


class SvsReader(DaPyramidChannelReader):

    def __init__(
        self,
        path: str | pathlib.Path,
        pixel_size: float | None = None
    ) -> None:
        # FIXME maybe move napari_lazy_openslide to optional dependency?
        # https://python-poetry.org/docs/pyproject/#extras
        # https://github.com/AllenCellModeling/aicsimageio/blob/main/aicsimageio/readers/bioformats_reader.py#L33-L40
        from napari_lazy_openslide import OpenSlideStore
       
        self.path = pathlib.Path(path)
        self.store = OpenSlideStore(str(self.path))
        self.zarr = zarr.open(self.store, mode='r')
        self._pixel_size = pixel_size
        pyramid = self.pyramid_from_svs()
        channel_axis = 2
        super().__init__(pyramid, channel_axis)

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['pyramid'], state['store'], state['zarr']
        state['path'] = state['path'].resolve()
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.__init__(path=state['path'], pixel_size=state['_pixel_size'])

    def pyramid_from_svs(self) -> list[da.Array]:
        return [
            da.from_zarr(self.store, component=d['path'])[..., :3]
            for d in self.zarr.attrs['multiscales'][0]['datasets']
        ]
   
    @property
    def pixel_size(self):
        if self._pixel_size is not None:
            return self._pixel_size
        try:
            return float(self.store._slide.properties['openslide.mpp-x'])
        except Exception:
            logger.warning(
                f'Unable to parse pixel size from {self.path.name};'
                f' assuming 1 µm. Use `_pixel_size` to set it manually'
            )
            self._pixel_size = 1
            return self._pixel_size



try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata

__version__ = importlib_metadata.version(__name__)


from . import (
    reader,
    align,
    align_multi_res,
    align_multi_obj,
    pyramid,
    color,
    
    # debugging
    block_affine,
    extract,
    img_util,
    register,
    register_util,
)



import math

import cv2
import dask.array as da
import numpy as np
import tifffile
import tqdm
import zarr
from loguru import logger

from . import __version__ as _version


class PyramidSetting:

    def __init__(
        self,
        downscale_factor=2,
        tile_size=1024,
        max_pyramid_img_size=1024
    ):
        self.downscale_factor = downscale_factor
        self.tile_size = tile_size
        self.max_pyramid_img_size = max_pyramid_img_size

    def tile_shapes(self, base_shape):
        shapes = np.array(self.pyramid_shapes(base_shape))
        n_rows_n_cols = np.ceil(shapes / self.tile_size)
        tile_shapes = np.ceil(shapes / n_rows_n_cols / 16) * 16
        return [tuple(map(int, s)) for s in tile_shapes]

    def pyramid_shapes(self, base_shape):
        num_levels = self.num_levels(base_shape)
        factors = self.downscale_factor ** np.arange(num_levels)
        shapes = np.ceil(np.array(base_shape) / factors[:,None])
        return [tuple(map(int, s)) for s in shapes]

    def num_levels(self, base_shape):
        factor = max(base_shape) / self.max_pyramid_img_size
        return math.ceil(math.log(factor, self.downscale_factor)) + 1


def format_channel_names(num_channels_each_mosaic, channel_names):
    '''
    format_channel_names(
        [1, 2, 3, 4, 5], ['x', 'x', ['x'], ['x', 'x']]
    )
    >>> [
        'x_1',
        'x_2',
        'x_3',
        'x_4',
        'x_5',
        'x_6',
        'Mosaic 4_1',
        'Mosaic 4_2',
        'Mosaic 4_3',
        'Mosaic 4_4'
    ]
    '''
    matched_channel_names = []
    for idx, (n, c) in enumerate(
        zip(channel_names, num_channels_each_mosaic)
    ):
        nl = n
        if type(n) == str:
            nl = [n] * c
        if len(nl) == 1:
            nl = nl * c
        if len(nl) != c:
            nl = [f"Mosaic {idx+1}"] * c
        matched_channel_names.append(nl)
    return make_unique_str(
        [n for l in matched_channel_names for n in l]
    )


def make_unique_str(str_list):
    if len(set(str_list)) == len(str_list):
        return str_list
    else:
        max_length = max([len(s) for s in str_list])
        str_np = np.array(str_list, dtype=np.dtype(('U', max_length+10)))
        unique, counts = np.unique(str_np, return_counts=True)
        has_duplicate = unique[counts > 1]
        for n in has_duplicate:
            suffixes = [
                f"_{i}"
                for i in range(1, (str_np == n).sum()+1)
            ]
            str_np[str_np == n] = np.char.add(n, suffixes)
    return make_unique_str(list(str_np))


def normalize_mosaics(mosaics):
    dtypes = set(m.dtype for m in mosaics)
    if any([np.issubdtype(d, np.floating) for d in dtypes]):
        max_dtype = np.dtype(np.float32)
    else:
        max_dtype = max(dtypes)
    normalized = []
    for m in mosaics:
        assert m.ndim == 2 or m.ndim == 3
        if m.ndim == 2:
            m = m[np.newaxis, :]
        normalized.append(m.astype(max_dtype, copy=False))
    return normalized


def write_pyramid(
    mosaics,
    output_path,
    pixel_size=1,
    channel_names=None,
    verbose=True,
    downscale_factor=4,
    compression=None,
    is_mask=False,
    tile_size=None,
    save_RAM=False,
    kwargs_tifffile=None
):
    mosaics = normalize_mosaics(mosaics)
    ref_m = mosaics[0]
    path = output_path
    num_channels = count_num_channels(mosaics)
    base_shape = ref_m.shape[1:3]
    assert int(downscale_factor) == downscale_factor
    assert downscale_factor < min(base_shape)
    pyramid_setting = PyramidSetting(
        downscale_factor=int(downscale_factor),
        tile_size=max(ref_m.chunksize)
    )
    num_levels = pyramid_setting.num_levels(base_shape)
    tile_shapes = pyramid_setting.tile_shapes(base_shape)
    shapes = pyramid_setting.pyramid_shapes(base_shape)

    if tile_size is not None:
        assert tile_size % 16 == 0, (
            f"tile_size must be None or multiples of 16, not {tile_size}"
        )
        tile_shapes = [(tile_size, tile_size)] * num_levels

    dtype = ref_m.dtype

    software = f'palom {_version}'
    pixel_size = pixel_size
    metadata = {
        'Creator': software,
        'Pixels': {
            'PhysicalSizeX': pixel_size, 'PhysicalSizeXUnit': '\u00b5m',
            'PhysicalSizeY': pixel_size, 'PhysicalSizeYUnit': '\u00b5m'
        },
    }

    if channel_names is not None:
        num_channels_each_mosaic = [
            count_num_channels([m])
            for m in mosaics
        ]
        names = format_channel_names(num_channels_each_mosaic, channel_names)
        if len(names) == num_channels:
            metadata.update({
                'Channel': {'Name': names},
            })

    logger.info(f"Writing to {path}")
    with tifffile.TiffWriter(path, bigtiff=True) as tif:
        kwargs = dict(
            metadata=metadata,
            software=software,
            compression=compression
        )
        if kwargs_tifffile is None:
            kwargs_tifffile = {}
        tif.write(
            data=tile_from_combined_mosaics(
                mosaics, tile_shape=tile_shapes[0], save_RAM=save_RAM
            ),
            shape=(num_channels, *shapes[0]),
            subifds=int(num_levels - 1),
            dtype=dtype,
            tile=tile_shapes[0],
            **{**kwargs, **kwargs_tifffile}
        )
        logger.info('Generating pyramid')
        for level, (shape, tile_shape) in enumerate(
            zip(shapes[1:], tile_shapes[1:])
        ):
            if verbose:
                logger.info(f"    Level {level+1} ({shape[0]} x {shape[1]})")
            tif.write(
                data=tile_from_pyramid(
                    path,
                    num_channels,
                    tile_shape=tile_shape,
                    downscale_factor=downscale_factor,
                    level=level,
                    is_mask=is_mask,
                    save_RAM=save_RAM
                ),
                shape=(num_channels, *shape),
                subfiletype=1,
                dtype=dtype,
                tile=tile_shape,
                **{
                    **dict(compression=compression),
                    **kwargs_tifffile
                }
            )


def count_num_channels(imgs):
    for img in imgs:
        assert img.ndim == 2 or img.ndim == 3
    return sum([
        1 if img.ndim == 2 else img.shape[0]
        for img in imgs
    ])
   

def tile_from_combined_mosaics(mosaics, tile_shape, save_RAM=False):
    num_rows, num_cols = mosaics[0].shape[1:3]
    h, w = tile_shape
    n = len(mosaics)
    for idx, m in enumerate(mosaics):
        for cidx, c in enumerate(m):
            # the performance is heavily degraded without pre-computing the
            # mosaic channel
            with tqdm.dask.TqdmCallback(
                ascii=True,
                desc=(
                    f"Assembling mosaic {idx+1:2}/{n:2} (channel"
                    f" {cidx+1:2}/{m.shape[0]:2})"
                ),
            ):
                c = da_to_zarr(c) if save_RAM else c.compute()
            for y in range(0, num_rows, h):
                for x in range(0, num_cols, w):
                    yield np.array(c[y:y+h, x:x+w])
                    # yield m[y:y+h, x:x+w].copy().compute()
            c = None


def tile_from_pyramid(
    path,
    num_channels,
    tile_shape,
    downscale_factor=2,
    level=0,
    is_mask=False,
    save_RAM=False
):
    # workaround progress bar
    # https://forum.image.sc/t/tifffile-ome-tiff-generation-is-taking-too-much-ram/41865/26
    pbar = tqdm.tqdm(total=num_channels, ascii=True, desc=f'Processing channel')
    for c in range(num_channels):
        img = da.from_zarr(zarr.open(tifffile.imread(
            path, series=0, level=level, aszarr=True
        )))
        if img.ndim == 2:
            img = img.reshape(1, *img.shape)
        img = img[c]
        # read using key seems to generate a RAM spike
        # img = tifffile.imread(path, series=0, level=level, key=c)
        if not is_mask:
            img = img.map_blocks(
                cv2.blur,
                ksize=(downscale_factor, downscale_factor), anchor=(0, 0)
            )
        img = da_to_zarr(img) if save_RAM else img.compute()
        num_rows, num_columns = img.shape
        h, w = tile_shape
        h *= downscale_factor
        w *= downscale_factor
        last_c = range(num_channels)[-1]
        last_y = range(0, num_rows, h)[-1]
        last_x = range(0, num_columns, w)[-1]
        for y in range(0, num_rows, h):
            for x in range(0, num_columns, w):
                if (y == last_y) & (x == last_x):
                    pbar.update(1)
                    if c == last_c:
                        pbar.close()
                yield np.array(img[y:y+h:downscale_factor, x:x+w:downscale_factor])
        # setting img to None seems necessary to prevent RAM spike
        img = None


def da_to_zarr(da_img, zarr_store=None, num_workers=None, out_shape=None, chunks=None):
    if zarr_store is None:
        if out_shape is None:
            out_shape = da_img.shape
        if chunks is None:
            chunks = da_img.chunksize
        zarr_store = zarr.create(
            out_shape,
            chunks=chunks,
            dtype=da_img.dtype,
            overwrite=True
        )
    da_img.to_zarr(zarr_store, compute=False).compute(
        num_workers=num_workers
    )
    return zarr_store